/*
 * Copyright (C) 2007 Richard Hughes <richard@hughsie.com>
 *
 * Licensed under the GNU Lesser General Public License Version 2
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <stdio.h>
#include <sys/stat.h>
#include <time.h>
#include <glib/gstdio.h>
#include <glib/gprintf.h>

#include <gmodule.h>
#include <glib.h>
#include <ohm-plugin.h>
#include <dbus/dbus-glib.h>

#define	HAL_DBUS_SERVICE		 	"org.freedesktop.Hal"
#define	HAL_DBUS_INTERFACE_POWER	 	"org.freedesktop.Hal.Device.SystemPowerManagement"
#define HAL_ROOT_COMPUTER		 	"/org/freedesktop/Hal/devices/computer"

enum {
	CONF_BUTTON_POWER_CHANGED,
	CONF_BUTTON_LID_CHANGED,
	CONF_TIMEOUTS_POWERDOWN_CHANGED,
	CONF_INHIBIT_CHANGED,
	CONF_LAST
};

static gboolean
set_dcon_sleep(OhmPlugin *plugin, gboolean sleep)
{
	FILE *fp;
	
	g_debug("in %s", G_STRFUNC);

	/* Turn off the backlight when sleeping.  We can't do this with
	   DPMS, because it would wake us back up on a keystroke. */ 
        
	fp = g_fopen("/sys/class/backlight/dcon-bl/device/sleep", "w");
	if (!fp) {
		g_warning ("Couldn't open /sys/class/backlight/dcon-bl/device/sleep");
		return FALSE;
	}

	if (g_fprintf(fp, sleep ? "1" : "0") < 0) {
		g_warning ("Couldn't write to /sys/class/backlight/dcon-bl/device/sleep.");
		return FALSE;
	}

	fclose(fp);
	return TRUE;
}
	
static gboolean
check_model(OhmPlugin *plugin)
{
	char model[5];
	FILE *fp;
	
	fp = g_fopen("/ofw/model", "r");
	if (!fp) {
		g_warning ("Couldn't open /ofw/model.");
		return FALSE;
	}

	while (fgets(model, sizeof(model), fp) != NULL) {
		g_debug ("Model is %s.", model);
	}

	if (g_ascii_strncasecmp(model, "C", 1) == 0) {
		g_debug ("C-series, returning TRUE.");
		return TRUE;
	}
	else {
		g_debug ("Not C-series, returning FALSE.");
		return FALSE;
	}
}

static gboolean
do_suspend(OhmPlugin *plugin)
{
	FILE *fp;

	g_debug("in %s", G_STRFUNC);
	fp = g_fopen("/sys/power/state", "w");
	if (!fp) {
		g_warning ("Couldn't open /sys/power/state.");
		return FALSE;
	}

	if (g_fprintf(fp, "mem") < 0) {
		g_warning ("Couldn't write to /sys/power/state.");
		return FALSE;
	}

	fclose(fp);

	g_debug("should have suspended");
	return TRUE;
}

static gboolean
pre_sleep(OhmPlugin *plugin)
{
	g_debug("in %s", G_STRFUNC);

	/* Turn off the backlight when sleeping.  We can't do this with
	   DPMS, because it would wake us back up on a keystroke. */ 

	set_dcon_sleep(plugin, TRUE);
	
	return TRUE;
}

static gboolean
pre_suspend(OhmPlugin *plugin)
{
  	g_debug("in %s", G_STRFUNC);
        return TRUE;
}

static void 
reset_idle(OhmPlugin *plugin)
{
	gint new_timeout;
	g_debug("in %s", G_STRFUNC);
	
	/* We've resumed now.  Bring the backlight back immediately. */
	ohm_plugin_conf_set_key (plugin, "backlight.state", 1);
	ohm_plugin_conf_set_key (plugin, "idle.reset", 1);
	ohm_plugin_conf_set_key (plugin, "idle.state", 0);

	ohm_plugin_conf_get_key (plugin, "timeouts.timer_momentary", &new_timeout);
	ohm_plugin_conf_set_key (plugin, "idle.timeout", new_timeout);
	ohm_plugin_conf_set_key (plugin, "timeouts.momentary", 0);
	ohm_plugin_conf_set_key (plugin, "timeouts.powerdown", 0);
}

static gboolean
post_suspend(OhmPlugin *plugin)
{
	reset_idle(plugin);
	return TRUE;
}

static gboolean
post_sleep(OhmPlugin *plugin)
{
	char source[256];
	FILE *fp;

	g_debug("in %s", G_STRFUNC);
	
	fp = g_fopen("/sys/power/wakeup-source", "r");
	if (!fp) {
		g_warning ("Couldn't open /sys/power/wakeup-source.");
		return TRUE;
	}

	while (fgets(source, sizeof(source), fp) != NULL) {
		g_debug ("Source was %s.", source);
	}

	fclose(fp);
	
	if ((g_ascii_strncasecmp(source, "power button", 12) == 0) ||
	    (g_ascii_strncasecmp(source, "lid", 3) == 0)) {
		/* Wake up fully. */
		/* FIXME: Don't wake if the lid is being *closed*. */
		g_debug("Waking up fully.");
		set_dcon_sleep(plugin, FALSE);
	}
	else {
		/* Go back to sleep. */
		g_debug("Going back to sleep.");
		/* X DPMS might have woken up the screen if we woke
		  due to a keypress. */
		set_dcon_sleep(plugin, TRUE);
		do_suspend(plugin);
		post_sleep(plugin);
	}

	reset_idle(plugin);
	return TRUE;
}

static void
power_button_pressed (OhmPlugin *plugin)
{
	gint value;

	gboolean supports_suspend;
	supports_suspend = check_model(plugin);
	if (!supports_suspend)
		return;

	/* check system inhibit - this is broken as any client can unref all */
	ohm_plugin_conf_get_key (plugin, "suspend.fixme_inhibit", &value);
	if (value == 1) {
		g_warning ("not doing lid closed action as inhibited");
		return;
	}

	pre_sleep(plugin);
	do_suspend(plugin);
	g_debug("Back in power_button_pressed.");
	post_sleep(plugin);
}

static void
lid_closed (OhmPlugin *plugin, gboolean is_closed)
{
	gint value;

	gboolean supports_suspend;
	supports_suspend = check_model(plugin);
	if (!supports_suspend)
		return;

	if (is_closed == FALSE) {
		return;
	}

	/* check system inhibit - this is broken as any client can unref all */
	ohm_plugin_conf_get_key (plugin, "suspend.fixme_inhibit", &value);
	if (value == 1) {
		g_warning ("not doing lid closed action as inhibited");
		return;
	}

	pre_sleep(plugin);
	do_suspend(plugin);
	post_sleep(plugin);
}

static gboolean
tty_in_use (OhmPlugin *plugin)
{
#define TTY1 "/dev/tty1"
#define TTY2 "/dev/tty2"

	struct stat tty1;
	struct stat tty2;
	time_t curtime;

	g_debug("in %s", G_STRFUNC);

	curtime = time(NULL);
	g_debug ("time now is %lu", (unsigned long) curtime);
	
	if (stat(TTY1, &tty1) >= 0) {
		g_debug ("tty1 time is %lu", (unsigned long) tty1.st_mtime);
		if (curtime-tty1.st_mtime < 60) {
			g_debug("tty1 is in use; inhibiting suspend.");
			return TRUE;
		}
	}

	if (stat(TTY2, &tty2) >= 0) {
		g_debug ("tty2 time is %lu", (unsigned long) tty2.st_mtime);
		if (curtime-tty2.st_mtime < 60) {
			g_debug("tty2 is in use; inhibiting suspend.");
			return TRUE;
		}
	}

	return FALSE;
}	
	
static gboolean 
system_is_idle (OhmPlugin *plugin)
{
#define SUSPEND_TAG_FILE "/etc/ohm/inhibit-suspend"

	gint value;
	struct stat stat_buf;

	gboolean supports_suspend;
	supports_suspend = check_model(plugin);
	if (!supports_suspend)
		return FALSE;

	g_debug("in %s", G_STRFUNC);

	/* check for inhibit file */
	if (stat(SUSPEND_TAG_FILE, &stat_buf) >= 0) {
		/* it exists */
		g_debug("not suspending; inhibit file found.");
		return FALSE;
	}
	
	/* check system inhibit - this is broken as any client can unref all */
	ohm_plugin_conf_get_key (plugin, "suspend.fixme_inhibit", &value);

	if (value == 1) {
		g_debug ("not doing idle action as inhibited");
		return FALSE;
	}

	if (tty_in_use(plugin))
		return FALSE;

	pre_suspend(plugin);
	do_suspend(plugin);
	post_suspend(plugin);
	return TRUE;
}

/**
 * plugin_notify:
 * @plugin: This class instance
 *
 * Notify the plugin that a key marked with ohm_plugin_conf_interested ()
 * has it's value changed.
 * An enumerated numeric id rather than the key is returned for processing speed.
 */
static void
plugin_notify (OhmPlugin *plugin, gint id, gint value)
{
	gint inhibit_suspend = 0;

	if (id == CONF_BUTTON_LID_CHANGED) {
		if (value == 1) {
			lid_closed (plugin, TRUE);
		} else {
			lid_closed (plugin, FALSE);
		}
	} else if (id == CONF_BUTTON_POWER_CHANGED) {
		/* only match on button press, not release */
		if (value == 1) {
			power_button_pressed (plugin);
		}
	} else if (id == CONF_TIMEOUTS_POWERDOWN_CHANGED) {
		/* only match on idle, not reset */
		g_debug ("powerdown changed");
		if (value == 1) {
			g_debug ("powerdown is 1, suspending");
			ohm_plugin_conf_get_key (plugin, "suspend.fixme_inhibit", &inhibit_suspend);
			if (inhibit_suspend == 1) {
				g_debug ("inhibit_suspend is 1, resetting");
				ohm_plugin_conf_set_key (plugin, "suspend.fixme_inhibit", 0);
				reset_idle(plugin);	
				return;
			}
			ohm_plugin_conf_get_key (plugin, "suspend.fixme_inhibit", &inhibit_suspend);
			if (inhibit_suspend == 0) {
				g_debug ("inhibit_suspend was 0, going ahead");
				if (!system_is_idle (plugin)) {
					g_debug ("didn't suspend -- resetting idle");
					reset_idle(plugin);
					return;
				}
			}
		}
		else if (value == 0) {
			g_debug ("powerdown is 0");
		}
	}
}

static void
plugin_init(OhmPlugin *plugin)
{
	ohm_plugin_conf_set_key (plugin, "suspend.fixme_inhibit", 0);

}

OHM_PLUGIN_DESCRIPTION (
	"suspend",			/* description */
	"0.0.1",			/* version */
	"richard@hughsie.com",		/* author */
	OHM_LICENSE_LGPL,		/* license */
	plugin_init,				/* initialize */
	NULL,				/* destroy */
	plugin_notify			/* notify */
);

OHM_PLUGIN_SUGGESTS (
	"timeouts",
	"buttons"
);

OHM_PLUGIN_INTERESTED (
	{"button.power", CONF_BUTTON_POWER_CHANGED},
	{"button.lid", CONF_BUTTON_LID_CHANGED},
	{"timeouts.powerdown", CONF_TIMEOUTS_POWERDOWN_CHANGED},
	{"suspend.fixme_inhibit", CONF_INHIBIT_CHANGED}
);
